/* scsicmd.c: generic SCSI command code
 * Copyright (C) 1996 Michael K. Johnson
 * 
 * I can be reached via email at johnsonm@redhat.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 * 
 */

#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>

#include "scsicmd.h"
#include "debug.h"

static int scsi_debug_out = 0;

/* ------------------------------------------------------------------------ */


/* pass in file descriptor for debugging */
void scsi_init_debug(int f)
{
  scsi_debug_out = f;
}


/* ------------------------------------------------------------------------ */


/*
 * do_scsi_cmd sends a SCSI command.
 * Returns:
 *  -1 if unexpected result
 *  -2 if failed write
 *  -3 if failed read
 *  the response code otherwise , see /usr/include/asm/errno.h
 *
 * obuf may be NULL if we don't care about the response...
 */

int do_scsi_cmd(int      sfd,          /* scanner device fd */
		unsigned char *ibuf,   /* input buffer */
		unsigned int in_size,      /* input data size */
		unsigned char *obuf,   /* output buffer */
		unsigned int *out_size_ptr      /* size of output buffer */
		)
{
  int status = 0;
  int out_size=*out_size_ptr;

  struct
  {
    struct sg_header sgh;
    char data[SG_BIG_BUFF];
  } *req;

  req = alloca(sizeof(* req));
  memcpy(req->data, ibuf, in_size);

  req->sgh.pack_len = SCSI_OFF + in_size;
  req->sgh.reply_len = SCSI_OFF + out_size;
  req->sgh.twelve_byte = 0; /* we assume this isn't necessary */

  if (scsi_debug_out)
  { dump_send(req->data , in_size, scsi_debug_out); }

  /* send command */
  status = write(sfd, req, SCSI_OFF + in_size);
  if (status < 0 || status != SCSI_OFF + in_size )
  {
    fprintf(stderr, "write failed; CMD = 0x%x, result = 0x%x, errno = %d\n",
	    ibuf[SCSI_OFF], req->sgh.result, errno);
    perror("");
    return -2;
  }

  if (!obuf) { obuf = ibuf; }

  /* retrieve result */
  status = read(sfd, req, SCSI_OFF + out_size);

  if (!(status == SCSI_OFF + out_size))
  {
    fprintf(stderr, "read failed to complete\n");
    return -3;
  }

  if (scsi_debug_out && obuf)
    { dump_receive(req->data , out_size, req->sgh.result, scsi_debug_out); }

  memcpy(obuf,req->data,out_size);


  return req->sgh.result;
}

