/* ------------------------------------------------------------------------- */

/* umax-cli.c: command-line-interface for UMAX scanner-driver.
  
   Copyright (C) 1996-1997 Michael K. Johnson
   Copyright (C) 1997 Oliver Rauch
  
   VERSION 0.5.6

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

 */


/* ------------------------------- INCLUDES -------------------------------- */


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <signal.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <getopt.h>

/* Get into kernel source tree... */
/* first need to create a dummy version of this structure that scsi.h wants.
 * Won't hurt us, just lets the compiler continue... */
struct request { char dummy; };

#include "scsi/scsi.h"
#include "scsi/sg.h"

#include "scsicmd.h"     
#include "debug.h"
#include "umax-struct.h"



/* -------------------------------- MAX ------------------------------------ */

#define max(a,b) (((a)>(b))?(a):(b))

/* ------------------------------- TRY TO ---------------------------------- */


#define try_to(command, error) if(command) {fprintf(stderr, error);return -1;}


/* ------------------------- GLOBAL VARIABLES ------------------------------ */

int only_print_inquiry;

/* ------------------------------ USAGE ------------------------------------ */


void usage(char * errormessage)
{
  fprintf(stderr, "\n%s\n", errormessage);
  fprintf(stderr,
  "\nUsage: umax-cli -ldgr|-VHI [options] [devicename]\n"
  "  l = lineart,   d = dithered,   g = greyscale,      r = rgb\n"
  "  V = Version,   H = Help,       I = print Inquiry (no scan)\n");

  fprintf(stderr,
  "\n"
  "Options:\n"
  "  o filename = write to file instead of stdout, mandatory for 3-pass-scan\n"
  "  W   = warmup,          q   = quality calibration\n"
  "  p   = preview\n"
  "  s   = slow speed       i   = invert colors\n"
  "  v # = verbose, 0 (nothing) .. 4 (all), 5 (all + scsi-trace in /tmp)\n"
  "  x # = x resolution,    y # = y resolution, R # = Resolution (both)\n"
  "  w # = width in pixels, h # = height in pixels\n"
  "  X # = ULX (left x),    Y # = ULY (upper y) (1200 units per inch)\n"
  "  t # = threshold, for lineart mode (128)\n"
  "  c # = contrast   (128,cbhs-range)\n"
  "  b # = brightness (128,cbhs-range)\n"
  "  L # = highlight (cbhs-range, maximum=default)\n"
  "  S # = shadow    (cbhs-range, minimum=default)\n"
  "  A # = analog gamma correction g (1.0 <= g <= 2.0)\n"
  "  G # = digital gamma correction g\n"
  "  C # = calibration mode ( 0=ignore, 1=driver, 2=image-type )\n"
  "  --Lred #, --Lgreen #, --Lblue # for separate selection of highlight\n"
  "  --Sred #, --Sgreen #, --Sblue # for separate selection of shadow\n"
  "  --Ared #, --Agreen #, --Ablue # for separate selection of analog gamma\n"
  "  --Gred #, --Ggreen #, --Gblue # for separate selection of digital gamma\n"
  "  --Gload file and --GRload,--GGload,--GBload = load gamma-file\n"
  "  --ADF   = use automatic document feeder\n"
  "  --UTA   = use umax-transparency-adapter\n"
  "  --DOR   = double optical resolution\n"
  "  --smear = don't care of image smearing problem\n"
  "  --CLD # = calibration lamp density (1-255), 0=auto (default)\n"
  "  --SLD # = scan lamp density (1-255), 0=use calibration-value (default)\n"
  "  --ETC # = exposure time for calibration in micro sec\n"
  "  --ETS # = exposure time for scan in micro sec\n"
  "     ETC/ETS: #=0: autoexposure, #=-1: default values\n"
  "  --ETCred #, --ETCgreen #, --ETCblue # for separate selection of ETC\n"
  "  --ETSred #, --ETSgreen #, --ETSblue # for separate selection of ETS\n"
  "\n"
  "Options that may help, if scanner is unknown or does not work:\n"
  "Try on your own risk, read manpage and documentation first!\n"
  "  --continue    = continue although scanner is unknown\n"
  "  --WDBlen #    = set window descriptor block length\n"
  "\n"
  "[devicename] is optional, if missing, /dev/scanner and /dev/sg[a-f0-7]\n"
  "are searched in order.\n"
  "\n");
  exit(1);
}

/* -------------------------- EVALUATE OPTIONS ----------------------------- */


void evaluate_options(US *us, int argc, char **argv)
{
  char c;
  int option_index;
  static struct option long_options[] =
  {
    {"Lred",1,0,0},
    {"Lgreen",1,0,0},
    {"Lblue",1,0,0},
    {"Sred",1,0,0},
    {"Sgreen",1,0,0},
    {"Sblue",1,0,0},
    {"Gred",1,0,0},
    {"Ggreen",1,0,0},
    {"Ared",1,0,0},
    {"Agreen",1,0,0},
    {"Ablue",1,0,0},
    {"Gblue",1,0,0},
    {"Gload",1,0,0},
    {"GRload",1,0,0},
    {"GGload",1,0,0},
    {"GBload",1,0,0},
    {"ADF",0,0,0},
    {"UTA",0,0,0},
    {"CLD",1,0,0},
    {"SLD",1,0,0},
    {"ETC",1,0,0},
    {"ETS",1,0,0},
    {"ETCred",1,0,0},
    {"ETCgreen",1,0,0},
    {"ETCblue",1,0,0},
    {"ETSred",1,0,0},
    {"ETSgreen",1,0,0},
    {"ETSblue",1,0,0},
    {"DOR",0,0,0},
    {"smear",0,0,0},
    {"auto",0,0,0},
    {"continue",0,0,0},
    {"WDBlen",1,0,0},
    {0,0,0,0}
  };

  while ((c = getopt_long(argc, argv,
          "VHIWldgrpqsix:y:R:w:h:X:Y:b:t:c:o:v:G:A:L:S:C:",
	  long_options, &option_index )) != EOF)
  {
    switch (c)
    {
    case 0:
      switch (option_index)
      {
        case 0: /* --Lred */
          {
           int highlight = atoi(optarg);
            if (highlight < 1)   { highlight = 1; }
            if (highlight > 255) { highlight = 255; }
            if (highlight < us->shadow_r)
            { us->shadow_r = highlight-1; }
            us->highlight_r = highlight;
          }
          break;
	  
        case 1: /* --Lgreen */
          {
           int highlight = atoi(optarg);
            if (highlight < 1)   { highlight = 1; }
            if (highlight > 255) { highlight = 255; }
            if (highlight < us->shadow_g)
            { us->shadow_g = highlight-1; }
            us->highlight_g = highlight;
          }
          break;

        case 2: /* --Lblue */
          {
           int highlight = atoi(optarg);
            if (highlight < 1)   { highlight = 1; }
            if (highlight > 255) { highlight = 255; }
            if (highlight < us->shadow_b)
            { us->shadow_b = highlight-1; }
            us->highlight_b = highlight;
          }
          break;

        case 3: /* --Sred */
          {
           int shadow = atoi(optarg);
            if (shadow < 0)   { shadow = 0; }
            if (shadow > 254) { shadow = 254; }
            if (us->highlight_r < shadow)
            { us->highlight_r = shadow+1; }
            us->shadow_r = shadow;
          }
          break;
	  
        case 4: /* --Sgreen */
          {
           int shadow = atoi(optarg);
            if (shadow < 0)   { shadow = 0; }
            if (shadow > 254) { shadow = 254; }
            if (us->highlight_g < shadow)
            { us->highlight_g = shadow+1; }
            us->shadow_g = shadow;
          }
          break;

        case 5: /* --Sblue */
          {
           int shadow = atoi(optarg);
            if (shadow < 0)   { shadow = 0; }
            if (shadow > 254) { shadow = 254; }
            if (us->highlight_b < shadow)
            { us->highlight_b = shadow+1; }
            us->shadow_b = shadow;
          }
          break;

        case 6: /* --Ared */
            us->analog_gamma_r = umax_calculate_analog_gamma(atof(optarg));
          break;
	  
        case 7: /* --Agreen */
            us->analog_gamma_g = umax_calculate_analog_gamma(atof(optarg));
          break;

        case 8: /* --Ablue */
            us->analog_gamma_b = umax_calculate_analog_gamma(atof(optarg));
          break;
	  
        case 9: /* --Gred */
	    us->gamma_r=atof(optarg);
	    us->digital_gamma_r=0;
          break;
	  
        case 10: /* --Ggreen */
	    us->gamma_g=atof(optarg);
	    us->digital_gamma_g=0;
          break;
	  
        case 11: /* --Gblue */
	    us->gamma_b=atof(optarg);
	    us->digital_gamma_b=0;
          break;

        case 12: /* Gload */
	    us->digital_gamma_r=us->digital_gamma_g=us->digital_gamma_b=-1;
	    us->gamma_r=us->gamma_g=us->gamma_b=0;
	    us->gamma_file_r=us->gamma_file_g=us->gamma_file_b=optarg;
          break;
	  
        case 13: /* GRload */
	    us->digital_gamma_r=-1;                       /* load gamma-file */
	    us->gamma_file_r=optarg;                         /* set filename */
	    us->gamma_r=0;
          break;
	  
        case 14: /* GGload */
	    us->digital_gamma_g=-1;                       /* load gamma-file */
	    us->gamma_file_g=optarg;                         /* set filename */
	    us->gamma_g=0;
          break;

        case 15: /* GBload */
	    us->digital_gamma_b=-1;                       /* load gamma-file */
	    us->gamma_file_b=optarg;                         /* set filename */
	    us->gamma_b=0;
          break;

        case 16: /* ADF */
          us->adf = 1;
          break;
	  
        case 17: /* UTA */
          us->uta = 1;
          break;

        case 18: /* CLD (calibration lamp density) */
          us->c_density = atoi(optarg);
          break;

        case 19: /* SLD (scan lamp density) */
          us->s_density = atoi(optarg);
          break;

        case 20: /* ETC (exposure time for calibration) */
          us->exposure_time_calibration_r = 
          us->exposure_time_calibration_g = 
          us->exposure_time_calibration_b = atoi(optarg);
          break;

        case 21: /* ETS (exposure time for scan) */
          us->exposure_time_calibration_r = 
          us->exposure_time_calibration_g = 
          us->exposure_time_calibration_b = atoi(optarg);
          break;

        case 22: /* ETCred (exposure time for calibration red) */
          us->exposure_time_calibration_r = atoi(optarg);
          break;

        case 23: /* ETCgreen (exposure time for calibration green) */
          us->exposure_time_calibration_g = atoi(optarg);
          break;

        case 24: /* ETCblue (exposure time for calibration blue) */
          us->exposure_time_calibration_b = atoi(optarg);
          break;

        case 25: /* ETSred (exposure time for scan red) */
          us->exposure_time_scan_r = atoi(optarg);
          break;

        case 26: /* ETSgreen (exposure time for scan green) */
          us->exposure_time_scan_g = atoi(optarg);
          break;

        case 27: /* ETSred (exposure time for scan red) */
          us->exposure_time_scan_b = atoi(optarg);
          break;

        case 28: /* DOR = double optical resolution */
          us->dor = 1;
          break;

        case 29: /* smear */
          us->smear = 1;
          break;

        case 30: /* auto  */
          us->set_auto = 1;
          break;

        case 31: /* continue (if scanner is unknown) */
          us->cont = 1;
          break;

        case 32: /* WDBlen # */
          us->wdb_len = atoi(optarg);
          break;
      }
      break;

    case 'H':
      usage("Umax help:");
    case 'V':
      fprintf(stderr,"Umax driver version 0.5.6\n");
    case 'W':
      us->warmup = 1;
      break;
    case 'l':
      us->colormode = LINEART;
      break;
    case 'd':
      us->colormode = HALFTONE;
      break;
    case 'g':
      us->colormode = GREYSCALE;
      break;
    case 'r':
      us->colormode = RGB;
      break;
    case 'v': /* verbose mode set by number now */
      us->verbose = max(0,atoi(optarg));
      break;
    case 's': /* speed=slow */
      us->slow = 1;
      break;
    case 'q':
      us->quality = 1;                                
      break;
    case 'i':
      us->reverse = 1;
      us->reverse_multi = 1;
      break;
    case 'p':
      us->preview = 1;                                
      break;
    case 'I':
      only_print_inquiry = 1;
      break;
    case 'x':
      us->x_resolution = max(0,atoi(optarg));
      break;
    case 'y':
      us->y_resolution = max(0,atoi(optarg));
      break;
    case 'R':
      us->x_resolution = us->y_resolution = max(0,atoi(optarg));
      break;
    case 'w':
      us->width_in_pixels = max(0,atoi(optarg));
      break;
    case 'h':
      us->height_in_pixels = max(0,atoi(optarg));
      break;
    case 'X':
      us->upper_left_x = max(0,atoi(optarg));
      break;
    case 'Y':
      us->upper_left_y = max(0,atoi(optarg));
      break;
    case 'b':
      us->brightness = atoi(optarg);
      break;
    case 't':
      us->threshold = atoi(optarg);
      break;
    case 'c':
      us->contrast = atoi(optarg);
      break;

    case 'o':
      us->ofd = open(optarg, O_CREAT|O_TRUNC|O_RDWR,
                    S_IRUSR|S_IWUSR | S_IRGRP|S_IWGRP | S_IROTH|S_IWOTH);
      break;

    case 'G':
        us->gamma_r= us->gamma_g= us->gamma_b=atof(optarg);
        us->digital_gamma_r= us->digital_gamma_g= us->digital_gamma_b=0;
      break;

    case 'A': /* set analog gamma value */
      {
        us->analog_gamma_r =
        us->analog_gamma_g =
        us->analog_gamma_b = umax_calculate_analog_gamma(atof(optarg));
      }
      break;

    case 'L':                                                   /* highlight */
      {
      int highlight;
        highlight = atoi(optarg);
        if (highlight < 1)   { highlight = 1; }
        if (highlight > 255) { highlight = 255; }
        if (highlight < us->shadow_r)
        { us->shadow_r = us->shadow_g = us->shadow_b = highlight-1; }
        us->highlight_r = us->highlight_g = us->highlight_b = highlight;
      }
      break;
     
    case 'S':                                                      /* shadow */
      {
      int shadow;
        shadow = atoi(optarg);
        if (shadow < 0)   { shadow = 0; }
        if (shadow > 254) { shadow = 254; }
        if (us->highlight_r < shadow)
        { us->highlight_r = us->highlight_g = us->highlight_b = shadow+1; }
        us->shadow_r = us->shadow_g = us->shadow_b = shadow;
      }
      break;

    case 'C':
      {
      int calibration= atoi(optarg);
        if ( (calibration < 0) || (calibration >2) )
	{ calibration = 0; }
        us->calibration = calibration;
      }
      break;

    case 0x3f:
      fprintf(stderr, "WARNING: getopt complains about bad use of option :-(\n"
	              "Trying to continue anyway...\n");
      break;

    default:
      fprintf(stderr, "Unknown option '-%c' (ascii %02xh)\n", c, c);
      fprintf(stderr, "type umax-cli -H for help\n");
      exit(1);
    }
  }
  if(optind < argc) { us->devicename = argv[optind]; }
}


/* ================================ TRY ==================================== */


#define try(command, error) if(command) { fprintf(stderr, error); exit(1); }


/* =============================== MAIN ==================================== */


/* Should call functions in this order:
 * open_scanner
 * us.grab_scanner
 * us.do_scan
 * us.give_scanner
 *
 * us.do_scan should call:
 *   set_window_param
 *   start_scan
 *   read_data
 * (or equivalent depending on the scanner)
 */

int main (int argc, char **argv)
{
int debug_out;
US us;

  if(argc < 2)
  {
    fprintf(stderr,"umax: too few arguments\n"
                   "type umax-cli -H for help\n");
    exit(1);
  }

  us.verbose=0;

  umax_init(&us);                         /* set fildescriptors and buffers */
  umax_initialize_values(&us);                     /* set values to default */

  only_print_inquiry = 0;
  evaluate_options(&us, argc, argv);   /* evaluate all command line options */

  if ((us.colormode==0) && (only_print_inquiry==0)) /* nothing selected? */
  {
    fprintf(stderr,
         "umax: choose (l)ineart, (d)ithered, (g)reyscale or (r)gb\n"
         "type umax-cli -H for help\n");
    exit(1);
  }

  if (us.verbose > 4)                    /* option -v 5 => enable scsi-trace */
  {
    debug_out = open("/tmp/umax.scsi.debug", O_WRONLY|O_CREAT|O_APPEND,
                     S_IRUSR|S_IWUSR | S_IRGRP|S_IWGRP | S_IROTH|S_IWOTH);
    if (!debug_out) /* open failed */ 
    { fprintf(stderr,"umax: Couldn't open /tmp/umax.scsi.debug\n");exit(1); }
    scsi_init_debug(debug_out);
  }

  try (umax_open_scanner(&us), "Could not open scanner\n");  /* open scanner */

  umax_get_inquiry_values(&us);                /* read/evaluate inquiry_data */

  if (only_print_inquiry)
  {
     print_inquiry(&us);
     exit(0);                   /* scanner is not grabbed, so its ok to quit */
  }

  umax_bound_scanarea(&us);                      /* get and correct scanarea */

  if (umax_check_values(&us) != 0)
  { fprintf(stderr,"ERROR: invalid scan-values\n"); exit(1);} 

  if (us.verbose) {print_variables(&us);}            /* print used variables */

  try (umax_grab_scanner(&us), "Could not allocate scanner\n");
    /* from here we may not abort program, we have to give scanner first !!! */

  if ((us.digital_gamma_r <= 0) ||
      (us.digital_gamma_g <= 0) ||
      (us.digital_gamma_b <= 0) )
  { send_gamma_curve(&us); }                      /* set digital gamma_curve */

  try (umax_do_scan(&us), "Could not do scan\n");                    /* scan */

  if (umax_give_scanner(&us))
  { fprintf (stderr, "Failed to deallocate scanner\n"); }

  exit (0);
}
