/* ------------------------------------------------------------------------- */

/* umax-driver.c: upper-level-part of scanner driver for UMAX scanners.
  
   Copyright (C) 1996-1997 Michael K. Johnson
   Copyright (C) 1997 Oliver Rauch
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version
   2 of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public
   License along with this program ( thats a lie at the moment)
   if not, write to the Free Software Foundation, Inc.,
   675 Mass Ave, Cambridge, MA 02139, USA.
   
 */


/* ------------------------------- INCLUDES -------------------------------- */


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <signal.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>

#include "scsicmd.h"             /* must come first; umax-scsi.h requires it */
#include "umax-scsi.h"
#include "umax-scsidef.h"
#include "umax-scanner.h"
#include "umax-struct.h"
#include "debug.h"


/* ---------------------------- MATH-HELPERS ------------------------------- */


#define min(a,b) (((a)<(b))?(a):(b))
#define max(a,b) (((a)>(b))?(a):(b))
#define inrange(minimum, val, maximum) (min(maximum,max(val,minimum)))


/* ------------------------------ WRITE HEADER ----------------------------- */


int write_header(US *us)
{
  char typebuf[512];

  /* Now put out the appropriate header.
   * With the appropriate header, it looks like raw scanner data is
   * the appropriate format below, so no munging needs to take place.
   * If 30-bit format is added, it will have to be output in *ascii*
   * mode to use pnm utilities.  Yuck!
   */

  switch(us->colormode) {
   case LINEART:
    sprintf(typebuf, "P4\n"
                     "# pbm created by umax\n"
                     "# scanner %s %s %s\n"
                     "# resolution %d dpi x %d dpi\n"
		     "# threshold %d\n"
		     "# LINEART\n"
		     "%u %u\n",
	    us->vendor, us->product, us->version,
	    us->x_resolution, us->y_resolution,
	    us->threshold,
	    us->width_in_pixels, us->height_in_pixels);
    break;
   case HALFTONE:
    sprintf(typebuf, "P4\n"
                     "# pbm created by umax\n"
                     "# scanner %s %s %s\n"
                     "# resolution %d dpi x %d dpi\n"
		     "# brightness %d, contrast %d\n"
		     "# highlight %d, shadow %d\n"
		     "# HALFTONE\n"
		     "%u %u\n",
	    us->vendor, us->product, us->version,
	    us->x_resolution, us->y_resolution,
	    us->brightness, us->contrast,
	    us->highlight_r, us->shadow_r,
	    us->width_in_pixels, us->height_in_pixels);
    break;
   case GREYSCALE:
    sprintf(typebuf, "P5\n"
                     "# pgm created by umax\n"
                     "# scanner %s %s %s\n"
                     "# resolution %d dpi x %d dpi\n"
		     "# highlight %d, shadow %d\n"
		     "# internal digital gamma curve %1.4f\n"
		     "# analog gamma %1.4f\n"
		     "# GREYSCALE\n"
                     "%u %u\n%u\n",
	    us->vendor, us->product, us->version,
	    us->x_resolution, us->y_resolution,
	    us->highlight_r, us->shadow_r,
	    us->gamma_r,
	    umax_get_analog_gamma(us->analog_gamma_r),
	    us->width_in_pixels, us->height_in_pixels, us->max_value);
    break;
   case RGB_LINEART:
   case RGB_HALFTONE:
   case RGB:
    sprintf(typebuf, "P6\n"
                     "# ppm created by umax\n"
                     "# scanner %s %s %s\n"
                     "# resolution %d dpi x %d dpi\n"
                     "# highlight r=%d g=%d b=%d\n"
                     "# shadow r=%d g=%d b=%d\n"
		     "# internal digital gamma curve r=%1.4f g=%1.4f b=%1.4f\n"
		     "# analog gamma r=%1.4f g=%1.4f b=%1.4f\n"
		     "# RGB\n"
                     "%u %u\n%u\n",
	    us->vendor, us->product, us->version,
	    us->x_resolution, us->y_resolution,
	    us->highlight_r, us->highlight_g, us->highlight_b,
	    us->shadow_r, us->shadow_g, us->shadow_b,
	    us->gamma_r, us->gamma_g, us->gamma_b,
	    umax_get_analog_gamma(us->analog_gamma_r),
	    umax_get_analog_gamma(us->analog_gamma_g),
	    umax_get_analog_gamma(us->analog_gamma_b),
	    us->width_in_pixels, us->height_in_pixels, us->max_value);
    break;
  }
  write(us->ofd, typebuf, strlen(typebuf));
  return 0;
}


/* ----------------------------- READ DATA --------------------------------- */


int read_data(US *us)
{
unsigned int data_left = 0;
unsigned int data_to_read;
unsigned int status;

  DBG(3,"starting to READ data\n");
  switch(us->colormode)
  {
   case LINEART:
    /* one bit per pixel */
    data_left = (us->width_in_pixels * us->height_in_pixels / 8);
    break;
    
   case HALFTONE: /* don't know if this is right ! */
    /* one bit per pixel */
    data_left = (us->width_in_pixels * us->height_in_pixels / 8);
    break;
    
   case GREYSCALE:
    /* one byte per pixel */
    data_left = us->width_in_pixels * us->height_in_pixels;
    break;

   case RGB_LINEART:
   case RGB_HALFTONE:
    data_left = us->width_in_pixels * us->height_in_pixels * 3 / 8;
    break;

   case RGB:
    /* three bytes per pixel.  Note that high-end scanners
     * might have 6, if they are 30 bits deep total, since
     * they spread each pixel over two bytes.  No weird packing!
     * If you have one of these scanners, please work with me
     * to extend this program to deal with your extra bits.
     */
    data_left = us->width_in_pixels * us->height_in_pixels * 3;
    break;
  }

  us->row_bufsize = us->bufsize;
  umax_trim_rowbufsize(us);                                 /* trim bufsize */
  DBG(3,"reading %u bytes in blocks of %u bytes\n",
        data_left, us->row_bufsize);

  do
  {
    /* if we want to, we can check to see if data_left agrees
     * with amount returned by GET DATA BUFFER STATUS, but why
     * bother? */

    data_to_read = (data_left < us->row_bufsize) ? data_left : us->row_bufsize;

    status = umax_read_data_block(us,data_to_read);
    if (status == 0) { continue; }
    if (status == -1) { return(-1); }

    /* output this chunk of data, which just happens to be
     * in pnm data format.  Yay! */
    write(us->ofd, us->buffer, data_to_read);

    data_left -= data_to_read;
    DBG(4, "buffer of %d bytes read; %d bytes to go\n",
           data_to_read, data_left);
  } while (data_left);

  DBG(3,"finished reading data\n");

  return 0;
}


/* -------------------- WRITE THREE PASS HEADER (PNM) ---------------------- */


int write_three_pass_header(US *us)
{
  char typebuf[256];
  int buflen = 0;

    if (us->arena)
    {
      if (us->three_pass_color != 1) { return 0; }
      buflen = sprintf(typebuf, "P6\n# ppm created by umax\n%u %u\n%u\n",
                   us->width_in_pixels, us->height_in_pixels, us->max_value);

      strncpy(us->arena, typebuf, buflen);
      us->arena_offset = buflen;
    }
    else
    {
      rewind(us->ofile);
      fprintf(us->ofile, "P6\n# ppm created by umax\n%u %u\n%u\n",
                   us->width_in_pixels, us->height_in_pixels, us->max_value);
      if (us->three_pass_color > 1)
      { fseek(us->ofile, us->three_pass_color-1, SEEK_CUR); }
    }

 return 0;
}


/* ---------------------- WRITE THREE PASS DATA  --------------------------- */


int write_three_pass_data(US *us, char *data, int len, int where)
{
  int i;

  if (us->arena)
  {
    for (i=0; i<len; i++)
    {
      *((us->arena+us->arena_offset) + (where*3) + 
        (us->three_pass_color-1)+(i*3)) = data[i];
    }
  }
  else
  {
    for (i=0; i<len; i++)
    {
      fputc(data[i], us->ofile);
      fseek(us->ofile, 2, SEEK_CUR);
    }
  }
 return 0;
}


/* -------------------------- READ THREE PASS DATA ------------------------- */


int read_three_pass_data(US *us)
{
  unsigned int data_left = 0;
  unsigned int data_read = 0;
  unsigned int data_to_read;
  unsigned int status;

  DBG(3, "starting to READ data (three pass)\n");

  data_left = us->width_in_pixels * us->height_in_pixels;

  us->row_bufsize = us->bufsize;
  umax_trim_rowbufsize(us);
  DBG(3, "reading %u bytes in blocks of %u bytes\n",
         data_left, us->row_bufsize);

  do
  {
    data_to_read = (data_left < us->row_bufsize) ? data_left : us->row_bufsize;

    status = umax_read_data_block(us,data_to_read);
    if (status == 0) { continue; }
    if (status == -1) { return(-1); }

    write_three_pass_data(us, us->buffer, data_to_read, data_read);

    data_left -= data_to_read;
    data_read += data_to_read;
    DBG(4, "buffer of %d bytes read; %d bytes to go\n",
           data_to_read, data_left);
  } while (data_left);

  DBG(3,"finished reading data pass %d\n", us->three_pass_color);

  return 0;
}

/* ------------------------------- TRY TO ---------------------------------- */


#define try_to(command, error) if(command) {DBG(1, error);return -1;}


/* ------------------------------ DO SCAN ---------------------------------- */


int umax_do_scan(US *us)
{
  /* for safty reposition scanner at first !!!!!! */
  /* that is for bad canceld scan last time */
  try_to(umax_reposition_scanner(us), "Could not reposition scan-unit\n");

  if ( (us->colormode != RGB ) || (us->three_pass == 0) ) /* singlepass scan */
  {
    try_to(umax_set_window_param(us), "Could not initialize scanner\n")
    try_to(umax_start_scan(us), "Could not start scan\n");
    try_to(write_header(us), "Could not write header\n");
    try_to(read_data(us), "Failed while attempting to read data\n");
    return 0;
  }
  else                                               /* color threepass scan */
  {
   int size = us->width_in_pixels * us->height_in_pixels * 3;
   int slop = 1024; /* for header */
   size_t pagesize;
   size_t mmap_size;
   int i;

    pagesize = getpagesize();
    mmap_size = size+slop + pagesize-((size+slop)%pagesize);
    ftruncate(us->ofd, mmap_size);

    us->arena = mmap(NULL, mmap_size, PROT_READ|PROT_WRITE,
                      MAP_FILE|MAP_SHARED, us->ofd, 0);

    if (us->arena == (void *) -1) { us->arena = 0; }
    if (us->arena == (void *) -1) { us->arena = NULL; }

    if (us->arena)
    { DBG(3, "reading data into memory-mapped file\n"); }
    else
    {
      DBG(3, "memory-mapping faild, reading data into file\n");
      us->ofile = fdopen(us->ofd,"w+");
    }

    for (i = 1; i <=3; i++)
    {
      us->three_pass_color=i;
      try_to(umax_set_window_param(us), "Could not initialize scanner\n");
      try_to(write_three_pass_header(us), "Could not write header\n");
      try_to(umax_start_scan(us), "Could not start three pass scan\n");
      try_to(read_three_pass_data(us),
             "Failed while attempting to read three pass data\n");
      try_to(umax_reposition_scanner(us), "Could not reposition scan-unit\n");
      /* reposition_scanner may possibly be removed, don't know! */
    }

    if (us->arena)
    {
      munmap(us->arena, mmap_size);
      DBG(3, "size = %d, offset = %d\n", size, us->arena_offset);
      ftruncate(us->ofd, size + us->arena_offset);
    }

   return 0;
  }
}



/* -------------------------- GENERATE GAMMA CURVE ------------------------- */


void generate_gamma_curve(char *field,double gamma)
{
int i;
  for(i=0;i<=255;i++)
  {
    *field = 255 * pow( ((double) i)/255.0, (1.0/gamma) );
    field++;
  }
}


/* ---------------------------- LOAD GAMMA CURVE --------------------------- */


void load_gamma_curve(char *field,char *filename)
{
int i,gamma_fd;
  gamma_fd=open(filename,O_RDONLY);
  if (gamma_fd == -1)
  {
    fprintf(stderr,"ERROR: Couldn't read gamma-curve-file %s\n",filename);
    exit(1);
  }
  if ( read(gamma_fd,field,256) <256)
  {
    fprintf(stderr,"ERROR: gamma-curve-file %s is to short\n",filename);
    close(gamma_fd);
    exit(1);
  }
}


/* ---------------------------- SEND GAMMA CURVE --------------------------- */


void send_gamma_curve(US *us)
{
int color;
char *dest;
char data[768];

  DBG(3,"uploading gamma curve\n");

  if (us->colormode == RGB) { color = 3; }
  else { color = 1; }

  dest = &data[0];
  if (us->digital_gamma_r == -1)               /* load or generate red gamma */
  {
    load_gamma_curve(dest,us->gamma_file_r);
    us->digital_gamma_r = 0;                   /* use downloaded gamma curve */
  }
  else
  { generate_gamma_curve(dest,us->gamma_r); }

  dest+=256;

  if (us->colormode == RGB)
  {
    if (us->digital_gamma_g == -1)            /* load or generate green data */
    {
      load_gamma_curve(dest,us->gamma_file_g);
      us->digital_gamma_g = 0;                 /* use downloaded gamma curve */
    }
    else
    { generate_gamma_curve(dest,us->gamma_g); }

    dest+=256;

    if (us->digital_gamma_b == -1)             /* load or generate blue data */
    {
      load_gamma_curve(dest,us->gamma_file_b);
      us->digital_gamma_b = 0;                 /* use downloaded gamma curve */
    }
    else
    { generate_gamma_curve(dest,us->gamma_b); }
  }

  umax_send_gamma_data(us, &data[0], color);
}

